#vider la mémoire
rm(list=ls())

#changer de répertoire
setwd("D:/DataMining/Databases_for_mining/dataset_for_soft_dev_and_comparison/optim_function_R")

#charger les données
library(xlsx)
D <- read.xlsx("calcul_avec_optim_R.xlsx",sheetIndex=1)
print(str(D))

#régression logistique
lr <- glm(coeur ~ age+taux_max+angine,data=D,family=binomial)

#coefficients
print(lr$coefficients)

#log-vraisemblance
print(-lr$deviance/2.0)

#matrice de var.covar de glm
rightVCOV <- vcov(lr)
print(rightVCOV)

#structures intermédiaires
endog <- D[,1]
exog <- as.matrix(D[,2:5])

#fonction pour initialisation du vecteur de paramètres
initialiser <- function(seed=1,ncol=4){
  #initialisation du générateur
  set.seed(seed)
  #valeurs de départ des coefficients
  coef <- runif(ncol)-0.5
  return(coef)
}

#fonction pour le calcul de la log-vraisemblance
myLL <- function(param,Y,X){
  #calcul de la combinaison linéaire, puis proba (pi)
  #pour chaque individu de la matrice X
  proba <- apply(X,1,function(x){v <- sum(param*x);return(1.0/(1.0+exp(-v)))})
  #log-vraisemblance
  LL <- sum((Y*log(proba)+(1.0-Y)*log(1.0-proba)))
  return(LL)
}

#initialisation
coef_depart <- initialiser()

#calcul de la vraisemblance avec ces coeffs
print(myLL(coef_depart,endog,exog))

#paramètres pour l'optimisation
#fnscale = -1 ==> maximisation
settings <- list(fnscale=-1)

#optimisation
res1 <- stats::optim(par=coef_depart,fn=myLL,method="BFGS",Y=endog,X=exog,control=settings)
print(res1)

#fonction pour calcul du gradient
myGradient <- function(param,Y,X){
  #calcul de la combinaison linéaire, puis pi (proba)
  #pour chaque individu de la matrice X
  proba <- apply(X,1,function(ligne){v <- sum(param*ligne);return(1.0/(1.0+exp(-v)))})
  #calcul du vecteur gradient
  gradient <- apply(X,2,function(colonne){sum((Y-proba)*colonne)})
  return(gradient)
}

#valeur du gradient à l'optimum GLM
gradGlm <- myGradient(lr$coefficients,endog,exog)
print(gradGlm)

#norme du vecteur gradient
print(sqrt(sum(gradGlm^2)))

#valeur du gradient à l'optimum res1
gradRes1 <- myGradient(res1$par,endog,exog)
print(gradRes1)

#sa norme
print(sqrt(sum(gradRes1^2)))

#départ vecteur de paramètres
coef_depart <- initialiser()

#optimisation
res2 <- stats::optim(par=coef_depart,fn=myLL,gr=myGradient,method="BFGS",Y=endog,X=exog,control=settings)
print(res2)

#vérification à la solution
gradRes2 <- myGradient(res2$par,endog,exog)
print(gradRes2)

#norme
print(sqrt(sum(gradRes2^2)))

#approximation de la matrice hessienne
Happrox <- optimHess(par=lr$coefficients,fn=myLL,Y=endog,X=exog)
print(Happrox)

#calcul de la matrice Hessienne
myHessian <- function(param,Y,X){
  #calcul de la combinaison linéaire, puis pi (proba)
  #pour chaque individu de la matrice X
  proba <- apply(X,1,function(x){v <- sum(param*x);return(1.0/(1.0+exp(-v)))})
  #matrice W -- diagonale avec proba*(1-proba)
  W <- matrix(0,nrow=nrow(X),ncol=nrow(X))
  diag(W) <- proba*(1.0-proba)
  #matrice Hessienne
  H <- t(X)%*%W%*%X
  return(-1.0*H)
}

#matrice hessienne avec la bonne formule
Hright <- myHessian(lr$coefficients,Y=endog,X=exog)
print(Hright)

#écart
ecartH <- sum((Happrox-Hright)^2)
print(ecartH)

#algorithme de newton-raphson
newton_raphson <- function(par,Y,X,maxiter=100){
  #moyennes des variables
  moy <- colMeans(X[,2:ncol(X)])
  #écarts-type
  et <- apply(X[,2:ncol(X)],2,sd)
  #centrer et réduire, pas la constante bien sûr
  Z <- cbind(X[,1],scale(X[,2:ncol(X)],center=moy,scale=et))
  #coefficient
  coef <- par
  #Log-vraisemblance
  LL <- myLL(coef,Y,Z)
  #compteur d'itérations
  iter <- 0
  #boucler jusqu'à convergence
  repeat{
    #compteur d'itération
    iter <- iter + 1
    #calcul du gradient
    grad <- myGradient(coef,Y,Z)
    #calcul de la matrice Hessienne
    hess <- myHessian(coef,Y,Z)
    #inversion de la matrice Hessienne
    invHess <- solve(hess)
    #màj des coefs
    newCoef <- coef - invHess%*%grad
    #recalcul de la LL
    newLL <- myLL(newCoef,Y,Z)
    #test de continuation
    #écart faible de l'évolution de la LL?
    diff <- (newLL - LL)
    if (diff < 1e-8 || iter > maxiter){
      #convergence - on sort de la boucle
      break  
    } else
    {
      #màj. coef
      coef <- newCoef
      #maj. LL
      LL <- newLL
    }
  }
  #recalculer les coefficients - dénormalisation
  #la constante
  coef[1] <- newCoef[1]+sum(sapply(1:3,function(j){-newCoef[j+1]*moy[j]/et[j]}))
  #les autres coefs
  for (j in 2:ncol(X)){coef[j] <- newCoef[j]/et[j-1]}
  #renvoyer les résultats
  return(list(loglik=newLL,coefficients=coef,iterations=iter))
}

#valeur de départ des coefficients
coef_depart <- initialiser()

#lancer les calculs
resNR <- newton_raphson(coef_depart,endog,exog)
print(resNR)

#valeur du gradient
gradNR <- myGradient(resNR$coefficients,endog,exog)
print(gradNR)

#norme du gradient
print(sqrt(sum(gradNR^2)))
